/*
 * parallel_matrix.cpp
 *
 * Scott Schneider
 */

#include "factory.h"
#include "work_unit.h"

#include <iostream>
using namespace std;

const int N = 1000;

class naive_matmul: public loop_unit {

	private:
		double *one;
		double *two;
		double *result;

	public:
		void work_init(double* _one, double* _two, double* _result, tree_unit* parent)
		{
			loop_unit::work_init(parent);
			one = _one;
			two = _two;
			result = _result;
		}

		void work()
		{
			for (int i = loop_start; i < loop_stop; ++i) {
				for (int j = 0; j < N; ++j) {
					double sum = 0;
					for (int k = 0; k < N; ++k) {
						sum += one[i * N + k] * two[k * N + j];
					}
					result[i * N + j] = sum;
				}
			}
		}
};

int main(int argc, char** argv)
{
	factory<naive_matmul> multiplier(LOGICAL, LOCAL);
	double* one	= new double[N * N];
	double* two	= new double[N * N];
	double* result	= new double[N * N];

	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < N; ++j) {
			one[i * N + j] = two[i * N + j] = 3;
			result[i * N + j] = 0;
		}
	}

	multiplier.start_working();

	naive_matmul root;
	multiplier.spawn_for(0, N, STATIC, one, two, result, &root);
	multiplier.root_barrier(&root);

	multiplier.stop_working();

	// correctness check
	if (argc == 2) {
		double* check = new double[N * N];

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				double sum = 0;
				for (int k = 0; k < N; ++k) {
					sum += one[i * N + k] * two[k * N + j];
				}
				check[i * N + j] = sum;
			}
		}

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				if ((result[i * N + j] >= (check[i * N + j]+0.05)) || (result[i * N + j] <= (check[i * N + j]-0.05))) {
					cout	<< "result[" << i << "][" << j << "] = " << result[i * N + j] << endl
						<< " check[" << i << "][" << j << "] = " << check[i * N + j] << endl
						<< "Incorrect." << endl;
					exit(1);
				}
			}
		}
		cout << "Correct." << endl;
	}

	return 0;
}
