/*
 * loop.cpp
 *
 * Scott Schneider
 */

#include <iostream>
using namespace std;

#include <cmath>
#include "factory.h"
#include "work_unit.h"
#include "timing.h"

#define ADD(i,j,k)	((k) = (i)+(j))

int NUMBER_OF_WORK_UNITS;
const int SAMPLE_SIZE = 500;
int NUMBER_OF_NOPS_PER_THREAD;

factory<mixed_work>* tester = NULL;

class noop: public task_unit
{
	int a;

	public:
		void work_init(int _a, noop* parent)
		{
			a = _a;
			task_unit::work_init(parent);
		}

		void work()
		{
			for (int i = 0; i < NUMBER_OF_NOPS_PER_THREAD; i += 10) {

				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				asm volatile ("pause");
				// 10
			}
		}
};

int main(int argc, char** argv)
{
	if (argc < 3) {
		cout << "usage: loop <# of threads> <# of nops>" << endl;
		exit(1);
	}

	bindme_all_double(0, atoi(argv[1]));
	NUMBER_OF_WORK_UNITS = atoi(argv[1]);
	NUMBER_OF_NOPS_PER_THREAD = atoi(argv[2]) / NUMBER_OF_WORK_UNITS;

	tester = new factory<mixed_work>(LOGICAL, atoi(argv[1]), STEAL);
	tester->register_class<noop>();
	tester->start_working();

	unsigned long long* samples = new unsigned long long[SAMPLE_SIZE];

	unsigned long long sum = 0;
	for (int i = 0; i < SAMPLE_SIZE; ++i) {

		unsigned long long before = clock_cycles();

		noop root;
		for (int j = 0; j < NUMBER_OF_WORK_UNITS; ++j) {
			tester->spawn<noop>(1, &root);
		}
		tester->root_barrier(&root);

		unsigned long long after = clock_cycles();

		samples[i] = after - before;
		sum += after - before;
	}

	tester->stop_working();
	
	unsigned long long average = sum / SAMPLE_SIZE;
	unsigned long long variance = 0;
	for (int i = 0; i < SAMPLE_SIZE; ++i) {
		variance += pow((long double)average - samples[i], 2);	
	}
	variance /= SAMPLE_SIZE;

	std::cout	<< "avg:" << average << std::endl
		 	<< "std:" << sqrt((long double)variance) << std::endl;

	return 0;
}

