/*
 * Scott Schneider
 *
 * complex.cpp
 */

#include <iostream>
using namespace std;

#include "work_unit.h"
#include "factory.h"

const int N = 2048;

factory<mixed_work>* matrixops;

class begin_arithmetic: public loop_unit
{
	private:
		double* a11;
		double* a12;
		double* a21;
		double* a22;

		double* b11;
		double* b12;
		double* b21;
		double* b22;

		double* s1;
		double* s2;
		double* s3;
		double* s4;

		double* t1;
		double* t2;
		double* t3;
		double* t4;

		int matrix_size;
		int a_width;
		int b_width;
		int c_width;

	public:
		void work_init( double* _a11, double* _a12, double* _a21, double* _a22, 
				double* _b11, double* _b12, double* _b21, double* _b22,
				double* _s1, double* _s2, double* _s3, double* _s4,
				double* _t1, double* _t2, double* _t3, double* _t4,
				int ms, int aw, int bw, int cw, task_unit* parent)
		{
			loop_unit::work_init(parent);

			a11 = _a11;
			a12 = _a12;
			a21 = _a21;
			a22 = _a22;

			b11 = _b11;
			b12 = _b12;
			b21 = _b21;
			b22 = _b22;

			s1 = _s1;
			s2 = _s2;
			s3 = _s3;
			s4 = _s4;

			t1 = _t1;
			t2 = _t2;
			t3 = _t3;
			t4 = _t4;

			matrix_size = ms;
			a_width = aw;
			b_width = bw;
			c_width = cw;
		}

		void work()
		{
			for (int i = loop_start; i < loop_stop; ++i) {
				for (int j = 0; j < matrix_size; ++j) {
					s1[i * matrix_size + j] = a21[i * a_width + j] + a22[i * a_width + j];
					t1[i * matrix_size + j] = b12[i * b_width + j] - b11[i * b_width + j];
					s3[i * matrix_size + j] = a11[i * a_width + j] - a21[i * a_width + j];
					t3[i * matrix_size + j] = b22[i * b_width + j] - b12[i * b_width + j];

					s2[i * matrix_size + j] = s1[i * matrix_size + j] - a11[i * a_width + j];
					t2[i * matrix_size + j] = b22[i * b_width + j] - t1[i * matrix_size + j];
					s4[i * matrix_size + j] = a12[i * a_width + j] - s2[i * matrix_size + j];
					t4[i * matrix_size + j] = b21[i * b_width + j] - t2[i * matrix_size + j];
				}
			}
		}
};

class end_arithmetic: public loop_unit
{
	private:
		double* c11;
		double* c12;
		double* c21;
		double* c22;

		double* p1;
		double* p2;
		double* p3;
		double* p4;
		double* p5;
		double* p6;
		double* p7;

		double* u2;
		double* u3;
		double* u6;

		int matrix_size;
		int c_width;

	public:
		void work_init( double* _c11, double* _c12, double* _c21, double* _c22,
				double* _p1, double* _p2, double* _p3, double* _p4, double* _p5, double* _p6, double* _p7, 
				double* _u2, double* _u3, double* _u6,
				int ms, int cw, task_unit* parent)
		{
			loop_unit::work_init(parent);

			c11 = _c11;
			c12 = _c12;
			c21 = _c21;
			c22 = _c22;

			p1 = _p1;
			p2 = _p2;
			p3 = _p3;
			p4 = _p4;
			p5 = _p5;
			p6 = _p6;
			p7 = _p7;

			u2 = _u2;
			u3 = _u3;
			u6 = _u6;

			matrix_size = ms;
			c_width = cw;
		}

		void work()
		{
			for (int i = loop_start; i < loop_stop; ++i) {
				for (int j = 0; j < matrix_size; ++j) {
					u2[i * matrix_size + j] = p1[i * matrix_size + j] + p4[i * matrix_size + j];
					u3[i * matrix_size + j] = u2[i * matrix_size + j] + p5[i * matrix_size + j];
					u6[i * matrix_size + j] = u2[i * matrix_size + j] + p3[i * matrix_size + j];

					c11[i * c_width + j] = p1[i * matrix_size + j] + p2[i * matrix_size + j];
					c12[i * c_width + j] = u6[i * matrix_size + j] + p6[i * matrix_size + j];
					c21[i * c_width + j] = u3[i * matrix_size + j] + p7[i * matrix_size + j];
					c22[i * c_width + j] = u3[i * matrix_size + j] + p3[i * matrix_size + j];
				}
			}
		}
};

class strassen: public task_unit
{
	private:
		double* a;
		double* b;
		double* c;
		int matrix_size;
		int a_width;
		int b_width;
		int c_width;

	public:
		void work_init(double* _a, double* _b, double* _c, int size, int aw, int bw, int cw, task_unit* parent)
		{
			task_unit::work_init(parent);
			a = _a;
			b = _b;
			c = _c;
			matrix_size = size;
			a_width = aw;
			b_width = bw;
			c_width = cw;
		}

		void work()
		{
			// base case
			if (matrix_size <= 64) {
				for (int i = 0; i < matrix_size; ++i) {
					for (int j = 0; j < matrix_size; ++j) {
						double sum = 0;
						for (int k = 0; k < matrix_size; ++k) {
							sum += a[i * a_width + k] * b[k * b_width + j];
						}
						c[i * c_width + j] = sum;
					}
				}
			}
			else {
				const int quad_size = matrix_size >> 1;

				double* a11 = a;
				double* a12 = a + quad_size;
				double* a21 = a + (a_width * quad_size);
				double* a22 = a + (a_width * quad_size) + quad_size;
				
				double* b11 = b;
				double* b12 = b + quad_size;
				double* b21 = b + (b_width * quad_size);
				double* b22 = b + (b_width * quad_size) + quad_size;

				double* c11 = c;
				double* c12 = c + quad_size;
				double* c21 = c + (c_width * quad_size);
				double* c22 = c + (c_width * quad_size) + quad_size;

				double* start_heap = new double[quad_size * quad_size * 18];
				double* heap = start_heap;

				double* s1 = heap; heap += quad_size * quad_size;
				double* s2 = heap; heap += quad_size * quad_size;
				double* s3 = heap; heap += quad_size * quad_size;
				double* s4 = heap; heap += quad_size * quad_size;

				double* t1 = heap; heap += quad_size * quad_size;
				double* t2 = heap; heap += quad_size * quad_size;
				double* t3 = heap; heap += quad_size * quad_size;
				double* t4 = heap; heap += quad_size * quad_size;

				double* p1 = heap; heap += quad_size * quad_size;
				double* p2 = heap; heap += quad_size * quad_size;
				double* p3 = heap; heap += quad_size * quad_size;
				double* p4 = heap; heap += quad_size * quad_size;
				double* p5 = heap; heap += quad_size * quad_size;
				double* p6 = heap; heap += quad_size * quad_size;
				double* p7 = heap; heap += quad_size * quad_size;

				double* u2 = heap; heap += quad_size * quad_size;
				double* u3 = heap; heap += quad_size * quad_size;
				double* u6 = heap; heap += quad_size * quad_size;

				matrixops->spawn_for<begin_arithmetic>(	0, quad_size, STATIC,
									a11, a12, a21, a22,
									b11, b12, b21, b22,
									s1, s2, s3, s4,
									t1, t2, t3, t4,
									quad_size, a_width, b_width, c_width, this);
				matrixops->child_barrier(this);

				// multiplications
				matrixops->spawn<strassen>(a11, b11, p1, quad_size, a_width, b_width, quad_size, this);
				matrixops->spawn<strassen>(a12, b21, p2, quad_size, a_width, b_width, quad_size, this);
				matrixops->spawn<strassen>(s1, t1, p3, quad_size, quad_size, quad_size, quad_size, this);
				matrixops->spawn<strassen>(s2, t2, p4, quad_size, quad_size, quad_size, quad_size, this);
				matrixops->spawn<strassen>(s3, t3, p5, quad_size, quad_size, quad_size, quad_size, this);
				matrixops->spawn<strassen>(s4, b22, p6, quad_size, quad_size, b_width, quad_size, this);	
				matrixops->spawn<strassen>(a22, t4, p7, quad_size, a_width, quad_size, quad_size, this);
				matrixops->child_barrier(this);

				matrixops->spawn_for<end_arithmetic>(	0, quad_size, STATIC,
									c11, c12, c21, c22,
									p1, p2, p3, p4, p5, p6, p7,
									u2, u3, u6,
									quad_size, c_width, this);
				matrixops->child_barrier(this);

				delete start_heap;

				matrixops->work_end_barrier(this);
			}
		}
};

int main(int argc, char** argv)
{
	double* a = new double[N * N];
	double* b = new double[N * N];
	double* c = new double[N * N];

	if (argc < 2) {
		cout << "my_complex <num_threads> [check]" << endl;
		exit(2);
	}

	for (int i = 0; i < N; ++i) {
		for (int j = 0; j < N; ++j) {
			a[i * N + j] = b[i * N + j] = 3;
			c[i * N + j] = 0;
		}
	}

	matrixops = new factory<mixed_work>(LOGICAL, atoi(argv[1]), STEAL);
	matrixops->start_working();
	barrier_object root = matrixops->spawn<strassen>(a, b, c, N, N, N, N, (strassen*)NULL);
	matrixops->root_barrier(root);
	matrixops->stop_working();
	delete matrixops;

	if (argc == 3) {
		double* check = new double[N * N];

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				double sum = 0;
				for (int k = 0; k < N; ++k) {
					sum += a[i * N + k] * b[k * N + j];
				}
				check[i * N + j] = sum;
			}
		}

		for (int i = 0; i < N; ++i) {
			for (int j = 0; j < N; ++j) {
				if ((c[i * N + j] >= (check[i * N + j]+0.05)) || (c[i * N + j] <= (check[i * N + j]-0.05))) {
					cout	<< "    c[" << i << "][" << j << "] = " << c[i * N + j] << endl
						<< "check[" << i << "][" << j << "] = " << check[i * N + j] << endl
						<< "Incorrect." << endl;
					exit(1);
				}
			}
		}
		cout << "Correct." << endl;
	}

	return 0;
}
