/*
 * loop.cpp
 *
 * Scott Schneider
 */

#include <iostream>
using namespace std;

#include <cmath>
#include <omp.h>
#include "bindme.h"
#include "timing.h"

const int SAMPLE_SIZE = 500;

int main(int argc, char** argv)
{
	if (argc < 3) {
		cout << "usage: loop <# of threads> <# of nops>" << endl;
		exit(1);
	}

	const int NUMBER_OF_WORK_UNITS = atoi(argv[1]);
	omp_set_num_threads(NUMBER_OF_WORK_UNITS);
	const int NUMBER_OF_NOPS_PER_THREAD = atoi(argv[2]) / NUMBER_OF_WORK_UNITS;

	unsigned long long* samples = new unsigned long long[SAMPLE_SIZE];

	unsigned long long sum = 0;
	for (int i = 0; i < SAMPLE_SIZE; ++i) {

		unsigned long long before = 0;
		unsigned long long after = 0;

		#pragma intel omp parallel
		{
			bindme_all_double(omp_get_thread_num(), omp_get_num_threads());
			before = clock_cycles();
			#pragma intel omp taskq
			{
				for (int j = 0; j < NUMBER_OF_WORK_UNITS; ++j) {
					#pragma intel omp task
					{
						for (int k = 0; k < NUMBER_OF_NOPS_PER_THREAD; k += 10) {
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							asm volatile ("pause");
							// 10
						}
					}
				}
			}
			after = clock_cycles();
		}

		samples[i] = after - before;
		sum += after - before;
	}

	unsigned long long average = sum / SAMPLE_SIZE;
	unsigned long long variance = 0;
	for (int i = 0; i < SAMPLE_SIZE; ++i) {
		variance += pow((long double)average - samples[i], 2);	
	}
	variance /= SAMPLE_SIZE;

	std::cout	<< "avg:" << average << std::endl
		 	<< "std:" << sqrt((long double)variance) << std::endl;

	return 0;
}

