/*
 * test_slab.cpp
 *
 * Scott Schneider
 */

#include "slab_allocator.h"
#include "posix_thread.h"

#include <iostream>
using namespace std;

const int TOTAL = 10000000;
const int NUMTHREADS = 8;

typedef general_manager Manager;

struct dummy: public plain_unit
{
	int a;
	int b;
	int c;
	int d;

	void work_init() {}
	void work() {}
};

void simulate_work(Manager* cache, int cpu, int bound)
{
	int q = cpu;

	slab_object* reserve = new slab_object[bound];

	for (int i = 0; i < TOTAL; ++i) {

		if (i % bound == 0 && i != 0) {
			for (int j = 0; j < bound; ++j) {
				reserve[j].free();
			}
		}

		reserve[i % bound] = cache->allocate<dummy>(q);
	}
}

struct simulate_args
{
	void(*function)(Manager*, int, int);

	Manager*	_1;	// cache
	int			_2;	// cpu
	int			_3;	// bound
};

int main(int argc, char* argv[])
{
	Manager cache;
	posix_thread<simulate_args>* threads = new posix_thread<simulate_args>[NUMTHREADS];

	cache.init(NUMTHREADS);
	cache.register_class<dummy>();

	int bound = atoi(argv[1]);

	for (int cpu = 0; cpu < NUMTHREADS; ++cpu) {
		threads[cpu].create(simulate_work, &cache, cpu, bound);
	}

	for (int cpu = 0; cpu < NUMTHREADS; ++cpu) {
		threads[cpu].join();
	}

	return 0;
}

