/*
 * bindme.h
 *
 * Scott Schneider, modified from code by Dimitrios Nikolopoulos.
 */

#ifndef __BINDME_H__
#define __BINDME_H__

#include <iostream>
#include "affinity.h"

const int MAX_CPUS = 8 * sizeof(cpu_set_t);

inline int bindme_cpu(int cpu)
{
	cpu_set_t set;

	CPU_ZERO(&set);
	CPU_SET(cpu, &set);
	int ret = sched_setaffinity(0, sizeof(cpu_set_t), &set);

	if (ret < 0) {

		std::cerr << "errno = " << errno << std::endl;
		if (errno == EFAULT) {
			std::cerr << "errno = EFAULT";
		}
		else if (errno == EINVAL) {
			std::cerr << "errno = EINVAL";
		}
		else if (errno == ESRCH) {
			std::cerr << "errno = ESRCH";
		}
		else if (errno == EPERM) {
			std::cerr << "errno = EPERM";
		}

		exit(-1);
	}

	return ret;
}

inline int bindme_physical_cpu(int context)
{
	cpu_set_t set;
	int neighbor;

	if (context % 2 == 0) {
		neighbor = context + 1;
	}
	else {
		neighbor = context - 1;
	}
	
	CPU_ZERO(&set);
	CPU_SET(context, &set);
	CPU_SET(neighbor, &set);
	int ret = sched_setaffinity(0, sizeof(cpu_set_t), &set);

	if (ret < 0) {

		std::cerr << "errno = " << errno << std::endl;
		if (errno == EFAULT) {
			std::cerr << "errno = EFAULT";
		}
		else if (errno == EINVAL) {
			std::cerr << "errno = EINVAL";
		}
		else if (errno == ESRCH) {
			std::cerr << "errno = ESRCH";
		}
		else if (errno == EPERM) {
			std::cerr << "errno = EPERM";
		}

		exit(-1);
	}

	return ret;
}

inline int bindme_all(int context, int num_threads)
{
	cpu_set_t set;

	CPU_ZERO(&set);
	
	for (int i = 0; i < num_threads; ++i) {
		CPU_SET(i, &set);
	}

	int ret = sched_setaffinity(0, sizeof(cpu_set_t), &set);

	if (ret < 0) {

		std::cerr << "errno = " << errno << std::endl;
		if (errno == EFAULT) {
			std::cerr << "errno = EFAULT";
		}
		else if (errno == EINVAL) {
			std::cerr << "errno = EINVAL";
		}
		else if (errno == ESRCH) {
			std::cerr << "errno = ESRCH";
		}
		else if (errno == EPERM) {
			std::cerr << "errno = EPERM";
		}

		exit(-1);
	}

	return ret;
}

inline int bindme_all_double(int context, int num_threads)
{
	cpu_set_t set;

	CPU_ZERO(&set);
	for (int i = 0; i < num_threads * 2; ++i) {
		CPU_SET(i, &set);
	}

	int ret = sched_setaffinity(0, sizeof(cpu_set_t), &set);

	if (ret < 0) {

		std::cerr << "errno = " << errno << std::endl;
		if (errno == EFAULT) {
			std::cerr << "errno = EFAULT";
		}
		else if (errno == EINVAL) {
			std::cerr << "errno = EINVAL";
		}
		else if (errno == ESRCH) {
			std::cerr << "errno = ESRCH";
		}
		else if (errno == EPERM) {
			std::cerr << "errno = EPERM";
		}

		exit(-1);
	}

	return ret;
}

static inline int get_num_cpus()
{
	return sysconf(_SC_NPROCESSORS_CONF);
}

#endif	// __BINDME_H__

