/*
 * Wrapper to ccmm.
 */

#ifndef __CCMM_ALLOCATOR_H__
#define __CCMM_ALLOCATOR_H__

extern "C" {
#include "../ccmm/ccmm.h"
}

class slab_object
{	
	friend class ccmm_allocator;

	private:
		work_unit*	object;

	public:
		slab_object(): object(NULL) {}
		slab_object(const slab_object& source): object(source.object) {}
		slab_object(work_unit* p): object(p) { }

		void operator=(work_unit* ptr)
		{
			object = ptr;
		}

		void operator=(const slab_object& source)
		{
			object = source.object;
		}

		int get_index()			{ return object->slab_index; }
		work_unit* get_object()		{ return object; }
		slab* get_slab()		{ return object->slab_ptr; }

		void set_index(int i)		{ object->slab_index = i; }
		void set_slab(slab* s)		{ object->slab_ptr = s; }

		bool operator==(const slab_object& s)	{ return object == s.object; }
		bool operator!=(const slab_object& s)	{ return object != s.object; }
		bool operator==(work_unit* ptr)		{ return object == ptr; }
		bool operator!=(work_unit* ptr)		{ return object != ptr; }
		work_unit* operator->() 		{ return object; }

		void free()
		{
			ccmm_free(object);
		}
};

class ccmm_allocator
{
	public:
		ccmm_allocator() {}
		~ccmm_allocator() {}

		void init(const unsigned int num_cpus) {}
		void local_init() { ccmm_thread_init(); }

		template <class T>
		void register_class() { }

		template <class T>
		slab_object allocate(const unsigned int q)
		{
			slab_object so = (work_unit*)ccmm_allocate(sizeof(T));
			new (so.object) T();
			return so;
		}
};


#endif	// __CCMM_ALLOCATOR_H__

