/*
 * clone_thread.h
 *
 * Scott Schneider
 */
 
#ifndef __CLONE_THREAD_H__
#define __CLONE_THREAD_H__

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <sched.h>

#include <linux/unistd.h>
//#include <unistd.h>

const int CLONESTACKSIZE = 16 *1024 * 1024; // 16MB

template <class Arguments>
class clone_thread
{
	private:
		pid_t	pid;

		template <class Function>
		void call_clone(Function f, Arguments* args)
		{
			void* sp;
			int ret;

			ret = posix_memalign(&sp, getpagesize(), CLONESTACKSIZE);
			if (ret != 0) {
				perror("clone_thread:call_clone::posix_memalign");
			}

			sp = (void*)((unsigned long)sp + CLONESTACKSIZE);

 			pid = clone((int(*)(void*))f, sp, SIGCHLD | CLONE_VM | CLONE_FS | CLONE_FILES, args);

			if (pid == -1) {
				perror("clone_thread::call_clone::clone");
			}
		}
	
		static int wrapper_1(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1);
			delete args;
			return 0;
		}

		static int wrapper_2(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2);
			delete args;
			return 0;
		}

		static int wrapper_3(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2, args->_3);
			delete args;
			return 0;
		}

		static int wrapper_4(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2, args->_3, args->_4);
			delete args;
			return 0;
		}

		static int wrapper_5(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2, args->_3, args->_4, args->_5);
			delete args;
			return 0;
		}

		static int wrapper_6(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2, args->_3, args->_4, args->_5, args->_6);
			delete args;
			return 0;
		}

		static int wrapper_7(Arguments* args)
		{
			#if defined(__GNUC__) && defined(__i386__)
			{
				(void)__builtin_alloca(16);
				 __asm__ __volatile__ ("andl $-16, %esp");

				#ifdef FFTW_DEBUG_ALIGNMENT
          			__asm__ __volatile__ ("addl $-4, %esp");
				#endif
     			}
			#endif

			#ifdef __ICC /* Intel's compiler for ia32 */
			{
				_alloca(16);
			}
			#endif

			args->function(args->_1, args->_2, args->_3, args->_4, args->_5, args->_6, args->_7);
			delete args;
			return 0;
		}

	public:
		clone_thread(): pid(-1) {}

		void join()
		{
			int ret;

			ret = waitpid(pid, NULL, 0); 
			if (ret == -1) {
				perror("clone_thread::join::waitpid");
			}
		}

		template <class Return>
		void create(Return (*f)())
		{
			call_clone<Return (*)(), void>(f, NULL);
		}
		
		template <class Return>
		void create(Return (*f)(Arguments*), Arguments* arg1)
		{
			call_clone(f, arg1);
		}

		template <class Function,
			class _1>
		void create(Function& f,
			_1 arg1)
		{
			Arguments temp = { f, arg1 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_1, args);
		}

		template <class Function,
			class _1,
			class _2>
		void create(Function f, 
			_1 arg1, 
			_2 arg2)
		{
			Arguments temp = { f, arg1, arg2 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_2, args);
		}

		template <class Function,
			class _1,
			class _2,
			class _3>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3)
		{
			Arguments temp = { f, arg1, arg2, arg3 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_3, args);
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
			_4 arg4)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_4, args);
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5>
		void create(Function f, 
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
		 	_4 arg4, 
			_5 arg5)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_5, args);
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5,
			class _6>
		void create(Function f,
			_1 arg1, 
			_2 arg2, 
			_3 arg3, 
			_4 arg4, 
			_5 arg5, 
			_6 arg6)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5, arg6 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_6, args);
		}

		template <class Function,
			class _1,
			class _2,
			class _3,
			class _4,
			class _5,
			class _6,
			class _7>
		void create(Function f,
			_1 arg1,
			_2 arg2,
			_3 arg3,
			_4 arg4,
			_5 arg5,
			_6 arg6,
			_7 arg7)
		{
			Arguments temp = { f, arg1, arg2, arg3, arg4, arg5, arg6, arg7 };
			Arguments* args = new Arguments(temp);
			call_clone(wrapper_7, args);
		}
};

#endif	// __CLONE_THREAD_H__

