/*
 * locks.h
 *
 * Scott Schneider. The atomic operations (in atomic.h) are taken
 * from the Nanos library.
 */

#ifndef __LOCKS_H__
#define __LOCKS_H__

#include "atomic.h"
#include "rw_lock.h"

const int CACHELINE = 128;

const unsigned long TAS_FREE = 0;
const unsigned long TAS_BUSY = 1;

/*
class spin_lock
{
	private:
		volatile unsigned long var;
		char padding[CACHELINE - sizeof(unsigned long)];

	public:
		spin_lock():
			var(TAS_FREE)
			{}

		void lock()
		{
			while (var == TAS_BUSY) {
				asm volatile ("pause");
			}

			while (fetch_and_store(&var, TAS_BUSY) == TAS_BUSY) {
				while (var == TAS_BUSY) {
					asm volatile ("pause");
				}
			}
		}

		void unlock()
		{
			var = TAS_FREE;
		}

};
*/

class exp_spin_lock
{
	private:
		volatile unsigned long var;
		char padding[CACHELINE - sizeof(unsigned long)];

	public:
		exp_spin_lock():
			var(TAS_FREE)
			{}

		void lock()
		{
			int backoff = 1;

			while (fetch_and_store(&var, TAS_BUSY) == TAS_BUSY) {
				for (int i = 0; i < backoff; ++i) {}
				if (backoff < 4096) {
					backoff *= 2;
				}
			}
		}

		void unlock()
		{
			var = TAS_FREE;
		}
};

class rw_lock
{
	private:
		volatile rw_lock_t var;
		char padding[CACHELINE - sizeof(rw_lock_t)];

	public:
		rw_lock()
		{
			rw_lock_init(&var);
		}

		void read_lock()
		{
			rw_lock_rdlock(&var);
		}

		void read_unlock()
		{
			rw_lock_rdunlock(&var);
		}

		void write_lock()
		{
			rw_lock_wrlock(&var);
		}

		void write_unlock()
		{
			rw_lock_wrunlock(&var);
		}

		void read_to_write_lock()
		{
			rw_lock_rdtowrlock(&var);
		}
};

class ticket_lock
{
	private:
		volatile unsigned long next_ticket;
		volatile unsigned long now_serving;
		char padding[CACHELINE - 2 * sizeof(unsigned long)];

	public:
		ticket_lock():
			next_ticket(0),
			now_serving(0)
			{}

		void lock()
		{
			volatile unsigned long my_ticket = atmc_fetch_and_add(&next_ticket, 1);

			do {
				unsigned long backoff = (my_ticket - now_serving) * 100;	
				for (unsigned long i = 0; i < backoff; ++i) {}

			} while (my_ticket != now_serving);
		}

		void unlock()
		{
			now_serving += 1;
		}
};

class noop_lock
{
	public:
		noop_lock() {}
		void lock() {}
		void unlock() {}
};

#endif	// __LOCKS_H__

