/*
 * Wrapper to C implementation of a lock free LIFO queue.
 */

#ifndef __QUEUE_WRAPPERS__
#define __QUEUE_WRAPPERS__

#include "c_queues.h"

class lf_lifo_queue
{
	private:
		lf_lifo_queue_t* q;
		slab_object buffer;

	public:
		lf_lifo_queue()		{ lf_lifo_queue_init(&q); }
		~lf_lifo_queue()	{ queue_end(q); }
		bool empty()		{ return q->top == NULL; }
		inline void pop()	{ }

		inline void push(slab_object& element)
		{
			int ret = lf_lifo_enqueue(q, reinterpret_cast<void*>(element.get_object()));
			assert(ret != 1);
		}

		inline slab_object& front()
		{
			void* obj = lf_lifo_dequeue(q);
			buffer = reinterpret_cast<work_unit*>(obj);
			return buffer;
		}
};

class lb_lifo_queue
{
	private:
		lb_lifo_queue_t* q;
		slab_object buffer;

	public:
		lb_lifo_queue()		{ lb_lifo_queue_init(&q); }
		~lb_lifo_queue()	{ queue_end(q); }
		bool empty()		{ return q->top == 0; }
		inline void pop()	{ }

		inline void push(slab_object& element)
		{
			int ret = lb_lifo_enqueue(q, reinterpret_cast<void*>(element.get_object()));
			assert(ret != 1);
		}

		inline slab_object& front()
		{
			void* obj = lb_lifo_dequeue(q);
			buffer = reinterpret_cast<work_unit*>(obj);
			return buffer;
		}
};

class lf_fifo_queue
{
	private:
		lf_fifo_queue_t* q;
		slab_object buffer;

	public:
		lf_fifo_queue()		{ lf_fifo_queue_init(&q); }
		~lf_fifo_queue()	{ queue_end(q); }
		bool empty()		{ return q->head == q->tail; }
		inline void pop()	{ }

		inline void push(slab_object& element)
		{
			int ret = lf_fifo_enqueue(q, reinterpret_cast<void*>(element.get_object()));
			assert(ret != 1);
		}

		inline slab_object& front()
		{
			void* obj = lf_fifo_dequeue(q);
			buffer = reinterpret_cast<work_unit*>(obj);
			return buffer;
		}
};

class lb_fifo_queue
{
	private:
		lb_fifo_queue_t* q;
		slab_object buffer;

	public:
		lb_fifo_queue()		{ lb_fifo_queue_init(&q); }
		~lb_fifo_queue()	{ queue_end(q); }
		bool empty()		{ return q->head == q->tail; }
		inline void pop()	{ }

		inline void push(slab_object& element)
		{
			int ret = lb_fifo_enqueue(q, reinterpret_cast<void*>(element.get_object()));
			assert(ret != 1);
		}

		inline slab_object& front()
		{
			void* obj = lb_fifo_dequeue(q);
			buffer = reinterpret_cast<work_unit*>(obj);
			return buffer;
		}
};

#endif	// __QUEUE_WRAPPERS__ 

