#ifndef __RW_SPIN_H_
#define __RW_SPIN_H_

/*
 * Do not change the values in the following definitions.
 * The implementation of the locking functions depends on
 * these exact values.
 */

typedef volatile unsigned long rw_lock_t;

#define RW_MAX_READERS	(0x7FFFFFFF)

/******************************************************************************/

static __inline__ void rw_lock_init(rw_lock_t *spin_var);
static __inline__ void rw_lock_rdlock(rw_lock_t *spin_var);
static __inline__ void rw_lock_wrlock(rw_lock_t *spin_var);
static __inline__ void rw_lock_rdtowrlock(rw_lock_t *spin_var);
static __inline__ void rw_lock_rdunlock(rw_lock_t *spin_var);
static __inline__ void rw_lock_wrunlock(rw_lock_t *spin_var);

/******************************************************************************/

static __inline__ void rw_lock_init(rw_lock_t *spin_var)
{
	*spin_var = RW_MAX_READERS;
}

/******************************************************************************/

static __inline__ void rw_lock_rdlock(rw_lock_t *spin_var)
{
	__asm__ __volatile__ (	"1:	lock		\n\t"
				"	decl	%0	\n\t"
				"	jns	3f	\n\t"
				"	lock		\n\t"
				"	incl	%0	\n\t"
				"2:	pause		\n\t"
				"	cmpl	$1, %0	\n\t"
				"	js	2b	\n\t"
				"	jmp	1b	\n\t"
				"3:			\n\t"
				: "=m" (*spin_var)
				:
				: "memory");
}

/******************************************************************************/

static __inline__ void rw_lock_wrlock(rw_lock_t *spin_var)
{
	__asm__ __volatile__ (	"	lock		\n\t"
				"	subl	%1, %0	\n\t"
				"	jz	2f	\n\t"
				"1:	pause		\n\t"
				"	cmpl	$0, %0	\n\t"
				"	jne	1b	\n\t"
				"2:			\n\t"
				: "=m" (*spin_var)
				: "i" (RW_MAX_READERS)
				: "memory");
}

/******************************************************************************/

static __inline__ void rw_lock_rdtowrlock(rw_lock_t *spin_var)
{
	__asm__ __volatile__ (	"	lock		\n\t"
				"	subl	%1, %0	\n\t"
				"	jz	2f	\n\t"
				"1:	pause		\n\t"
				"	cmpl	$0, %0	\n\t"
				"	jne	1b	\n\t"
				"2:			\n\t"
				: "=m" (*spin_var)
				: "i" (RW_MAX_READERS - 1)
				: "memory");
}

/******************************************************************************/

static __inline__ void rw_lock_rdunlock(rw_lock_t *spin_var)
{
	__asm__ __volatile__ (	"lock			\n\t"
				"incl	%0		\n\t"
				: "=m" (*spin_var)
				:
				: "memory");
}

/******************************************************************************/

static __inline__ void rw_lock_wrunlock(rw_lock_t *spin_var)
{
	__asm__ __volatile__ (	"lock			\n\t"
				"addl	%1, %0		\n\t"
				: "=m" (*spin_var)
				: "i" (RW_MAX_READERS)
				: "memory");
}

/******************************************************************************/

#endif   /* __RW_SPIN_H_ */
