/*
 * Scott Schneider
 *
 * simple_deque.h
 */

#ifndef __SIMPLE_DEQUE_H__
#define __SIMPLE_DEQUE_H__

#include <stack>
#include <iostream>
#include <cmath>

template <class T>
class std_lifo_queue
{
	private:
		std::stack<T> s;

	public:
		inline bool empty()		{ return s.empty(); }
		inline void push(const T& elem)	{ s.push(elem); }
		inline void pop()		{ s.pop(); }
		inline T& front()		{ return s.top(); }	// only reason this class is needed
};

template <class T>
class simple_deque_iterator
{
	private:
		T* data;
		int index;
		int absolute_size;

	public:
		simple_deque_iterator(): data(NULL), index(-1), absolute_size(-1) {}
		simple_deque_iterator(T* d, int i, int s): data(d), index(i), absolute_size(s) {}
		void operator++()	{ index = (index + 1) % absolute_size; }
		T& operator*()		{ return data[index]; }

		friend bool operator== <>(const simple_deque_iterator<T>& a, const simple_deque_iterator<T>& b);
		friend bool operator!= <>(const simple_deque_iterator<T>& a, const simple_deque_iterator<T>& b);
};

template <class T>
bool operator==(const simple_deque_iterator<T>& a, const simple_deque_iterator<T>& b)
{
	return a.index == b.index;
}

template <class T>
bool operator!=(const simple_deque_iterator<T>& a, const simple_deque_iterator<T>& b)
{
	return a.index != b.index;
}

const int STARTSIZE = 5;

template <class T>
class simple_deque
{
	private:	
		T* data;
		unsigned int absolute_size;
		unsigned int head;
		unsigned int tail;

	public:
		typedef simple_deque_iterator<T> iterator;

		simple_deque()
		{
			absolute_size = STARTSIZE;
			data = new T[absolute_size];
			tail = 0;
			head = 1;
		}

		~simple_deque()
		{
			delete [] data;
		}

		void push_front(const T& v)
		{
			if (tail == (head - 2) % absolute_size) {
				T* temp = new T[absolute_size * 2];
				for (unsigned int i = 0; i < absolute_size; ++i) {
					temp[i] = data[i];
				}
				delete [] data;
				data = temp;
				absolute_size *= 2;

			}
			head = (head - 1) % absolute_size;
			data[head] = v;
		}

		void push_back(const T& v)
		{
			if (head == (tail + 2) % absolute_size) {
				T* temp = new T[absolute_size * 2];
				for (unsigned int i = 0; i < absolute_size; ++i) {
					temp[i] = data[i];
				}
				delete [] data;
				data = temp;
				absolute_size *= 2;
			}
			tail = (tail + 1) % absolute_size;
			data[tail] = v;
		}

		void pop_front()
		{
			head = (head + 1) % absolute_size;
		}

		void pop_back()
		{
			tail = (tail - 1) % absolute_size;
		}

		unsigned int size() const
		{
			return ((tail + absolute_size) - head  + 1) % absolute_size;
		}

		void print()
		{
			for (unsigned int i = 0; i < absolute_size; ++i) {
				std::cout << i << ": " << data[i] << std::endl;
			}
		}

		void clear()
		{
			tail = 0;
			head = 1;
		}

		T& front()			{ return data[head]; }
		T& back()			{ return data[tail]; }
		iterator begin()		{ return iterator(data, tail, absolute_size); }
		iterator end()			{ return iterator(data, head, absolute_size); }
		bool empty() const		{ return size() == 0; }
};

#endif	// __SIMPLE_DEQUE_H__

