/*
 * Scott Schneider
 *
 * simple_list.h
 */

#ifndef __SIMPLE_LIST_H__
#define __SIMPLE_LIST_H__

#include <cstdlib>

template <class T>
class simple_list;

template <class T>
struct simple_node
{
	T data;
	simple_node* next;
	simple_node* previous;

	simple_node(const T& d, simple_node* n, simple_node* p): data(d), next(n), previous(p) {}
};

template <class T>
class simple_list_iterator
{
	friend class simple_list<T>;

	private:
		simple_node<T>* node;

	public:

		simple_list_iterator(): node(NULL) {}
		simple_list_iterator(simple_node<T>* n): node(n) {}
		void operator++()	{ node = node->next; }
		T& operator*()		{ return node->data; }

		friend bool operator== <>(const simple_list_iterator<T>& a, const simple_list_iterator<T>& b);
		friend bool operator!= <>(const simple_list_iterator<T>& a, const simple_list_iterator<T>& b);
};

template<class T>
bool operator==(const simple_list_iterator<T>& a, const simple_list_iterator<T>& b)
{
	return a.node == b.node;
}

template<class T>
bool operator!=(const simple_list_iterator<T>& a, const simple_list_iterator<T>& b)
{
	return a.node != b.node;
}

template <class T>
class simple_list
{
	private:
		simple_node<T>* head;
		simple_node<T>* tail;

	public:
		typedef simple_list_iterator<T> iterator;

		simple_list()
		{
			tail = new simple_node<T>(T(), NULL, NULL);
			head = tail;
		}

		~simple_list()
		{
			for (simple_node<T>* cur = head; cur != NULL; cur = cur->next) {
				if (cur->previous) {
					delete cur->previous; 
				}
			}
			delete tail;
		}

		void push_front(const T& v)
		{
			simple_node<T>* node = new simple_node<T>(v, head, NULL);
			head->previous = node;
			head = node;
		}

		void push_back(const T& v)
		{
			simple_node<T>* node = new simple_node<T>(v, tail, tail->previous);
			if (tail->previous != NULL) {
				tail->previous->next = node;
			}
			else {
				head = node;
			}
			tail->previous = node;
		}

		void splice(iterator end, simple_list<T>& x, iterator begin)
		{
			if (head != tail->previous) {
			simple_node<T>* temp = head;
			head = head->next;
			head->previous = NULL;

			temp->next = tail;
			temp->previous = tail->previous;
			if (tail->previous) {
				tail->previous->next = temp;
			}
			tail->previous = temp;
			}
		}

		T& front()		{ return head->data; }
		T& back()		{ return tail->previous->data; }
		iterator begin()	{ return iterator(head); }
		iterator end()		{ return iterator(tail); }
};

#endif	// __SIMPLE_LIST_H__

